"""App Settings"""

# Alliance Auth (External Libs)
from app_utils.app_settings import clean_setting

# Switch between AA-Corp Stats and CorpStats Two APP
KILLSTATS_CORPSTATS_TWO = clean_setting("KILLSTATS_CORPSTATS_TWO", False)

# Set Naming on Auth Hook
KILLSTATS_APP_NAME = clean_setting("KILLSTATS_APP_NAME", "Killstats")

# If True you need to set up the Logger
KILLSTATS_LOGGER_USE = clean_setting("KILLSTATS_LOGGER_USE", False)

# zKillboard - https://zkillboard.com/
ZKILLBOARD_BASE_URL = "https://zkillboard.com/"
ZKILLBOARD_API_URL = "https://zkillboard.com/api/"

ZKILLBOARD_BASE_URL_REGEX = r"^http[s]?:\/\/zkillboard\.com\/"
ZKILLBOARD_KILLMAIL_URL_REGEX = r"^http[s]?:\/\/zkillboard\.com\/kill\/\d+\/"

# Caching Key for Caching System
STORAGE_BASE_KEY = "killstats_storage_"

# Max lifetime of killmails in temporary storage in seconds
KILLBOARD_STORAGE_LIFETIME = clean_setting("KILLBOARD_STORAGE_LIFETIME", 3_600 * 1)

# Max Killmails per run should not be higher then 500
KILLBOARD_MAX_KILLMAILS_PER_RUN = clean_setting("KILLBOARD_MAX_KILLMAILS_PER_RUN", 400)

# Max lifetime of ZKB Request (30 min), Don't go lower then 1800 seconds
KILLBOARD_ZKB_CACHE_LIFETIME = clean_setting("KILLBOARD_ZKB_CACHE_LIFETIME", 1800 * 1)

# Tasks hard timeout
KILLBOARD_TASKS_TIMEOUT = clean_setting("KILLBOARD_TASKS_TIMEOUT", 1_800)



import sys
from unittest.mock import MagicMock, patch

from corpstats.models import CorpMember as ExpectedCorpMember

from django.core.exceptions import ObjectDoesNotExist
from django.test import RequestFactory, TestCase
from esi.models import Token

from allianceauth.corputils.models import CorpMember, CorpStats
from allianceauth.eveonline.models import EveCharacter, EveCorporationInfo
from app_utils.testing import create_user_from_evecharacter

from killstats.api.account_manager import AccountManager
from killstats.errors import KillstatsImportError
from killstats.tests.testdata.load_allianceauth import load_allianceauth

MODULE_PATH = "killstats.api.account_manager"


class TestApiHelpers(TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        load_allianceauth()
        cls.factory = RequestFactory()

        cls.user, cls.char_owner = create_user_from_evecharacter(
            1001,
            permissions=[
                "killstats.basic_access",
            ],
        )
        cls.user2, _ = create_user_from_evecharacter(
            1002,
            permissions=[
                "killstats.basic_access",
            ],
        )
        cls.user3, _ = create_user_from_evecharacter(
            1003,
        )
        cls.user4, _ = create_user_from_evecharacter(
            1004,
        )
        cls.user5, _ = create_user_from_evecharacter(
            1005,
        )
        cls.corp = EveCorporationInfo.objects.get(corporation_id=2001)

    def test_get_main_and_alts_all_char_in_chars(self):
        # given
        mains = {}
        request = self.factory.get("/")
        request.user = self.user
        corp_stats = CorpStats.objects.create(
            token=Token.objects.get(user=self.user),
            corp=self.corp,
        )
        CorpMember.objects.create(
            character_id=1005, character_name="Gerthd", corpstats=corp_stats
        )
        chars = EveCharacter.objects.filter(
            corporation_id__in=[self.corp.corporation_id]
        )
        for char in chars:
            mains[char.character_id] = {"main": char, "alts": [char]}
        excepted_data = mains
        account = AccountManager(corporations=[self.corp.corporation_id])
        # when
        data, _ = account.get_mains_alts()
        # then
        self.assertEqual(data, excepted_data)

    @patch(MODULE_PATH + ".EveCharacter.objects.select_related")
    def test_get_main_and_alts_all_process_corpmember(self, mock_select_related):
        # given
        mock_char = MagicMock()
        mock_char.character_id = 1005
        mock_char.character_name = "Gerthd"
        mock_select_related.return_value.get.return_value = mock_char

        request = self.factory.get("/")
        request.user = self.user
        corp_stats = CorpStats.objects.create(
            token=Token.objects.get(user=self.user),
            corp=self.corp,
        )
        CorpMember.objects.create(
            character_id=9999, character_name="Test9999", corpstats=corp_stats
        )
        account = AccountManager(corporations=[self.corp.corporation_id])
        # when
        data, _ = account.get_mains_alts()
        # then
        mock_select_related.assert_called()
        self.assertIn("MagicMock", str(data.values()))

    @patch(MODULE_PATH + ".EveCharacter.objects.select_related")
    def test_get_main_and_alts_all_process_corpmember_object_does_not_exist(
        self, mock_select_related
    ):
        # Setup mock EveCharacter instance to simulate ObjectDoesNotExist
        mock_select_related.return_value.get.side_effect = ObjectDoesNotExist

        # Setup the rest of your test environment as before
        mains = {}
        request = self.factory.get("/")
        request.user = self.user
        corp_stats = CorpStats.objects.create(
            token=Token.objects.get(user=self.user),
            corp=self.corp,
        )
        CorpMember.objects.create(
            character_id=9999, character_name="Test9999", corpstats=corp_stats
        )
        chars = EveCharacter.objects.filter(
            corporation_id__in=[self.corp.corporation_id]
        )
        for char in chars:
            mains[char.character_id] = {"main": char, "alts": [char]}
        account = AccountManager(corporations=[self.corp.corporation_id])
        # when
        data, _ = account.get_mains_alts()

    @patch(MODULE_PATH + ".app_settings.KILLSTATS_CORPSTATS_TWO", True)
    def test_get_corp_models_and_string(self):
        account = AccountManager()
        CorpMember = account.get_corp_models_and_string()
        self.assertIs(CorpMember, ExpectedCorpMember)

    def test_process_character(self):
        # Create mock main character
        main_char = EveCharacter(character_id=1002, corporation_id=2002)

        # Mock char with a main character
        char = EveCharacter(character_id=1001, corporation_id=2001)
        char.character_ownership = self.char_owner
        char.character_ownership.user.profile.main_character = main_char

        # Test main not in characters
        characters = {}
        chars_list = set()
        missing_chars = set()

        account = AccountManager(corporations=[2001])

        account._process_character(char, characters, chars_list, missing_chars)

        # Test main in characters
        characters = {1002: {"main": main_char, "alts": []}}
        chars_list = set()
        missing_chars = set()

        account._process_character(char, characters, chars_list, missing_chars)

        # Test Corporation exist
        characters = {}
        chars_list = set()
        missing_chars = set()

        account._process_character(char, characters, chars_list, missing_chars)

        # Test Corporation not exist
        char = EveCharacter(character_id=1001, corporation_id=2001)
        characters = {}
        chars_list = set()
        missing_chars = set()

        account._process_character(char, characters, chars_list, missing_chars)

        # Test Attribute Error
        error = EveCorporationInfo.objects.get(corporation_id=2001)
        characters = {}
        chars_list = set()
        missing_chars = set()

        account._process_character(error, characters, chars_list, missing_chars)

        # Test Missing character
        char = EveCharacter(character_id=9999, corporation_id=2001)
        characters = {}
        chars_list = set()
        missing_chars = set()

        account._process_character(char, characters, chars_list, missing_chars)

        # Test Corporation & Alliance Empty
        account = AccountManager()
        char = EveCharacter(character_id=1001, corporation_id=2001)
        characters = {}
        chars_list = set()
        missing_chars = set()

        account._process_character(char, characters, chars_list, missing_chars)


class TestApiHelperCorpStatsImport(TestCase):
    def setUp(self):
        self.original_sys_modules = sys.modules.copy()

    def tearDown(self):
        sys.modules = self.original_sys_modules

    @patch(MODULE_PATH + ".app_settings.KILLSTATS_CORPSTATS_TWO", True)
    @patch(MODULE_PATH + ".logger")
    def test_packages_are_not_installed(self, mock_logger):
        with patch.dict(
            sys.modules,
            {k: None for k in list(sys.modules) if k.startswith("corpstats")},
        ):
            with self.assertRaises(KillstatsImportError):
                account = AccountManager()
                _ = account.get_corp_models_and_string()
            mock_logger.error.assert_called_with(
                "Corpstats is enabled but not installed"
            )



"""API"""

# Standard Library
from typing import List

from ninja import NinjaAPI

# AA Killstats
from killstats.api import schema
from killstats.api.killboard.killboard_helper import (
    get_killmails_data,
    get_killstats_halls,
    get_killstats_stats,
)
from killstats.hooks import get_extension_logger
from killstats.models.killstatsaudit import AlliancesAudit

logger = get_extension_logger(__name__)


class KillboardAllianceApiEndpoints:
    tags = ["Killboard"]

    def __init__(self, api: NinjaAPI):
        # Killmails
        @api.get(
            "killmail/month/{month}/year/{year}/alliance/{alliance_id}/{mode}/",
            response={200: dict, 403: str},
            tags=self.tags,
        )
        def get_alliance_killmails(
            request, month, year, alliance_id: int, mode, page_size: int = 100
        ):
            return get_killmails_data(
                request, month, year, alliance_id, mode, page_size, "alliance"
            )

        # Hall of Fame/Shame
        @api.get(
            "halls/month/{month}/year/{year}/alliance/{alliance_id}/",
            response={200: List[schema.KillboardHall], 403: str},
            tags=self.tags,
        )
        def get_halls(request, month, year, alliance_id: int):
            return get_killstats_halls(request, month, year, alliance_id, "alliance")

        # Stats
        @api.get(
            "stats/month/{month}/year/{year}/alliance/{alliance_id}/",
            response={200: List[schema.KillboardStats], 403: str},
            tags=self.tags,
        )
        def get_killstats(request, month, year, alliance_id: int):
            return get_killstats_stats(request, month, year, alliance_id, "alliance")

        # Admin
        @api.get(
            "killboard/alliance/admin/",
            response={200: List[schema.AllianceAdmin], 403: str},
            tags=self.tags,
        )
        def get_corporation_admin(request):
            alliances = AlliancesAudit.objects.visible_to(request.user)

            if not alliances:
                return 403, "Permission Denied"

            alliance_dict = {}

            for alliance in alliances:
                # pylint: disable=broad-exception-caught
                try:
                    alliance_dict[alliance.alliance.alliance_id] = {
                        "alliance_id": alliance.alliance.alliance_id,
                        "alliance_name": alliance.alliance.alliance_name,
                    }
                except Exception:
                    continue

            output = []
            output.append({"alliance": alliance_dict})

            return output



"""App Views"""

from django.contrib import messages

# Django
from django.contrib.auth.decorators import login_required, permission_required
from django.shortcuts import redirect, render
from esi.decorators import token_required

from allianceauth.eveonline.models import (
    EveAllianceInfo,
    EveCharacter,
    EveCorporationInfo,
)

# AA Killstats
from killstats import __title__
from killstats.models.killstatsaudit import AlliancesAudit, CorporationsAudit
from killstats.tasks import killmail_update_corp

from .hooks import get_extension_logger

logger = get_extension_logger(__name__)


@login_required
@permission_required("killstats.basic_access")
def killboard_index(request, corporation_pk=0, alliance_pk=0):
    context = {
        "corporation_pk": corporation_pk,
        "alliance_pk": alliance_pk,
    }
    return render(request, "killstats/killboard.html", context=context)


@login_required
@token_required(scopes=("publicData"))
@permission_required(["killstats.admin_access"])
def add_corp(request, token):
    char = EveCharacter.objects.get_character_by_id(token.character_id)
    if char:
        corp, _ = EveCorporationInfo.objects.get_or_create(
            corporation_id=char.corporation_id,
            defaults={
                "member_count": 0,
                "corporation_ticker": char.corporation_ticker,
                "corporation_name": char.corporation_name,
            },
        )

        _, created = CorporationsAudit.objects.update_or_create(
            corporation=corp, owner=char
        )
        if created:
            killmail_update_corp.apply_async(args=[char.corporation_id], priority=6)
        msg = f"{char.corporation_name} successfully added/updated to Killboard"
        messages.info(request, msg)
        return redirect("killstats:index")

    msg = "Failed to add Corporation to Killboard"
    messages.error(request, msg)
    return redirect("killstats:index")


@login_required
@token_required(scopes=("publicData"))
@permission_required(["killstats.admin_access"])
def add_alliance(request, token):
    char = EveCharacter.objects.get_character_by_id(token.character_id)
    if char:
        try:
            alliance = EveAllianceInfo.objects.get(alliance_id=char.alliance_id)
        except EveAllianceInfo.DoesNotExist:
            alliance = EveAllianceInfo.objects.create_alliance(
                alliance_id=char.alliance_id
            )

        if alliance:
            _, created = AlliancesAudit.objects.update_or_create(
                alliance=alliance, owner=char
            )
            if created:
                pass
                # killmail_update_ally.apply_async(args=[char.alliance_id], priority=6)
            msg = f"{char.alliance_name} successfully added/updated to Killboard"
            messages.info(request, msg)
            return redirect("killstats:index")

    msg = "Failed to add Alliance to Killboard"
    messages.error(request, msg)
    return redirect("killstats:index")


@login_required
@permission_required("killstats.basic_access")
def corporation_admin(request):
    """
    Corporation Admin
    """
    context = {}
    return render(request, "killstats/admin/corporation_admin.html", context=context)


@login_required
@permission_required("killstats.basic_access")
def alliance_admin(request):
    """
    Alliance Admin
    """
    context = {}
    return render(request, "killstats/admin/alliance_admin.html", context=context)



from django.core.exceptions import ObjectDoesNotExist
from django.db.models import Q

from allianceauth.eveonline.models import EveCharacter

from killstats import app_settings
from killstats.errors import KillstatsImportError
from killstats.hooks import get_extension_logger

logger = get_extension_logger(__name__)


class AccountManager:
    def __init__(self, corporations=None, alliances=None) -> None:
        self.corporations = corporations if corporations is not None else []
        self.alliances = alliances if alliances is not None else []

    # pylint: disable=import-outside-toplevel
    def get_corp_models_and_string(self):
        if app_settings.KILLSTATS_CORPSTATS_TWO:
            try:
                from corpstats.models import CorpMember

                return CorpMember
            except ImportError as exc:
                logger.error("Corpstats is enabled but not installed")
                raise KillstatsImportError(
                    "Corpstats is enabled but not installed"
                ) from exc

        from allianceauth.corputils.models import CorpMember

        return CorpMember

    def _get_linked_characters(self):
        query_filter = Q()
        if self.corporations:
            query_filter = Q(corporation_id__in=self.corporations)
            query_filter |= Q(
                character_ownership__user__profile__main_character__corporation_id__in=self.corporations
            )
        else:
            query_filter = Q(alliance_id__in=self.alliances)
            query_filter |= Q(
                character_ownership__user__profile__main_character__alliance_id__in=self.alliances
            )

        linked_chars = EveCharacter.objects.filter(query_filter)
        return (
            linked_chars.select_related(
                "character_ownership",
                "character_ownership__user__profile__main_character",
            )
            .prefetch_related("character_ownership__user__character_ownerships")
            .order_by("character_name")
        )

    def _process_character(
        self, char: EveCharacter, characters: dict, chars_list: set, missing_chars: set
    ):
        try:
            main = char.character_ownership.user.profile.main_character
            if main and main.character_id not in characters:
                characters[main.character_id] = {"main": main, "alts": []}
            if (
                char.corporation_id in self.corporations
                or char.alliance_id in self.alliances
            ):
                characters[main.character_id]["alts"].append(char)
                chars_list.add(char.character_id)
        except ObjectDoesNotExist:
            if EveCharacter.objects.filter(character_id=char.character_id).exists():
                char = EveCharacter.objects.get(character_id=char.character_id)
                characters[char.character_id] = {"main": char, "alts": []}
                if (
                    char.corporation_id in self.corporations
                    or char.alliance_id in self.alliances
                ):
                    chars_list.add(char.character_id)
                    characters[char.character_id]["alts"].append(char)

            missing_chars.add(char.character_id)
        except AttributeError:
            pass

    def get_mains_alts(self):
        """Get all members for given corporations/alliances"""
        characters = {}
        chars_list = set()
        missing_chars = set()

        linked_chars = self._get_linked_characters()

        for char in linked_chars:
            self._process_character(char, characters, chars_list, missing_chars)

        if self.corporations:
            corpmember = self.get_corp_models_and_string()
            for member in corpmember.objects.filter(
                corpstats__corp__corporation_id__in=self.corporations
            ).exclude(character_id__in=chars_list):
                try:
                    char = EveCharacter.objects.select_related(
                        "character_ownership",
                        "character_ownership__user__profile__main_character",
                    ).get(character_id=member.character_id)
                    self._process_character(char, characters, chars_list, missing_chars)
                except ObjectDoesNotExist:
                    missing_chars.add(member.character_id)

        # TODO Maybe create task for missing_chars if needed

        return characters, list(chars_list)



# Standard Library
from collections import Counter
from heapq import nlargest

from eveuniverse.models import EveType

# Alliance Auth (External Libs)
from allianceauth.eveonline.evelinks import eveimageserver

# AA Killstats
from killstats.hooks import get_extension_logger
from killstats.models.general import EveEntity
from killstats.models.killboard import Killmail

logger = get_extension_logger(__name__)


# TODO Refactor this code and make it easier...


class KillboardStatsCount:
    def __init__(self):
        self.worst_ship = 0
        self.top_victim = 0
        self.top_ship = 0
        self.top_killer = 0
        self.alltime_killer = 0
        self.alltime_loss = 0


class KillboardStatsManager:
    def __init__(self, date, stats, entities):
        self.worst_ship = Counter()
        self.top_victim = Counter()
        self.top_ship = Counter()
        self.top_killer = Counter()
        self.alltime_killer = Counter()
        self.alltime_loss = Counter()
        self.count = KillboardStatsCount()
        self.entities = entities
        self.date = date
        self.stats = stats

    def update_for_killmail(self, killmail: Killmail):
        self._update_kill_stats(killmail)

    def _update_kill_stats(self, killmail: Killmail):
        counted_ships = set()
        for attacker in killmail.attackers:
            if (
                attacker["corporation_id"] in self.entities
                or attacker["alliance_id"] in self.entities
            ):
                ship_id = attacker["ship_type_id"]
                attacker_id = attacker["character_id"]
                if ship_id not in counted_ships and killmail.get_month(self.date.month):
                    self.top_ship[ship_id] += 1
                    counted_ships.add(ship_id)
                if killmail.get_month(self.date.month):
                    self.top_killer[attacker_id] += 1
                self.alltime_killer[attacker_id] += 1
        victim_id = killmail.victim.eve_id
        if (
            killmail.victim_corporation_id in self.entities
            or killmail.victim_alliance_id in self.entities
        ):
            if killmail.get_month(self.date.month):
                if not killmail.is_capsule() and not killmail.is_mobile():
                    self.worst_ship[killmail.victim_ship.id] += 1
                self.top_victim[victim_id] += 1
            self.alltime_loss[victim_id] += 1

    def _stats_count(self, title, model, count, char=False, loss=False):
        if model:
            eve_id = model.eve_id if char else model.id

            if char:
                portrait = eveimageserver.character_portrait_url(eve_id, 256)
            else:
                portrait = eveimageserver.type_icon_url(eve_id, 256)

            key_id = "character_id" if char else "ship"
            key_name = "character_name" if char else "ship_name"
            self.stats.append(
                {
                    "title": title,
                    "type": "count",
                    key_id: eve_id,
                    key_name: model.name,
                    "portrait": f"{portrait}",
                    "count": count,
                    "loss": loss,
                    "zkb_link": f"https://zkillboard.com/character/{eve_id}/",
                }
            )

    def get_top_entity(self, counter, model):
        try:
            top_entity_id, top_entity_count = counter.most_common(1)[0]
            if model == EveType:
                top_entity = model.objects.get(id=top_entity_id)
            else:
                top_entity = model.objects.get(eve_id=top_entity_id)
            return top_entity, top_entity_count
        except IndexError:
            return None, 0

    # Losses
    def update_worst_ship(self):
        self.worst_ship, self.count.worst_ship = self.get_top_entity(
            self.worst_ship, EveType
        )
        self._stats_count(
            "Worst Ship:",
            self.worst_ship,
            self.count.worst_ship,
            loss=True,
        )

    def update_top_victim(self):
        self.top_victim, self.count.top_victim = self.get_top_entity(
            self.top_victim, EveEntity
        )
        self._stats_count(
            f"Top Victim: {self.top_victim}",
            self.top_victim,
            self.count.top_victim,
            char=True,
            loss=True,
        )

    def update_alltime_loss(self):
        self.alltime_loss, self.count.alltime_loss = self.get_top_entity(
            self.alltime_loss, EveEntity
        )
        self._stats_count(
            f"Alltime Victim: {self.alltime_loss}",
            self.alltime_loss,
            self.count.alltime_loss,
            char=True,
            loss=True,
        )

    # Kills
    def update_top_ship(self):
        self.top_ship, self.count.top_ship = self.get_top_entity(self.top_ship, EveType)
        self._stats_count("Top Ship:", self.top_ship, self.count.top_ship)

    def update_top_killer(self):
        self.top_killer, self.count.top_killer = self.get_top_entity(
            self.top_killer, EveEntity
        )
        self._stats_count(
            f"Top Killer: {self.top_killer}",
            self.top_killer,
            self.count.top_killer,
            char=True,
        )

    def update_alltime_killer(self):
        self.alltime_killer, self.count.alltime_killer = self.get_top_entity(
            self.alltime_killer, EveEntity
        )
        self._stats_count(
            f"Alltime Killer: {self.alltime_killer}",
            self.alltime_killer,
            self.count.alltime_killer,
            char=True,
        )


def killboard_process_kills(killmail_data: Killmail, entities):
    killmail_data = killmail_data.prefetch_related("victim", "victim_ship")
    kills = []
    losses = []
    totalvalue = 0
    totalvalue_loss = 0

    for killmail in killmail_data.filter_entities_kills(entities):
        _stats_killmail(killmail, kills)
        totalvalue += killmail.victim_total_value

    for killmail in killmail_data.filter_entities_losses(entities):
        _stats_killmail(killmail, losses)
        totalvalue_loss += killmail.victim_total_value

    return kills, totalvalue, losses, totalvalue_loss


def _evulate_portrait(killmail, mode="killmail"):
    eve_id = killmail.victim.eve_id
    portrait = eveimageserver.character_portrait_url(eve_id, 256)
    if killmail.victim.category == "corporation":
        portrait = eveimageserver.corporation_logo_url(eve_id, 256)
    if killmail.victim.category == "alliance":
        portrait = eveimageserver.alliance_logo_url(eve_id, 256)
    if mode == "ship":
        portrait = (
            f"https://images.evetech.net/types/{killmail.victim_ship.id}/icon?size=256"
        )
    return portrait


def _evaluate_zkb_link(killmail):
    zkb = f"https://zkillboard.com/character/{killmail.victim.eve_id}/"
    if killmail.victim.category == "corporation":
        zkb = f"https://zkillboard.com/corporation/{killmail.victim.eve_id}/"
    if killmail.victim.category == "alliance":
        zkb = f"https://zkillboard.com/alliance/{killmail.victim.eve_id}/"
    return zkb


def _get_character_details(killmail, mains, mode):
    main = None
    main_id = None
    alt = None

    # Check if Victim is Main or Alt and set main and alt accordingly
    if killmail.victim.eve_id in mains:
        main_data = mains[killmail.victim.eve_id]
        main = main_data["main"]
        main_id = killmail.victim.eve_id
    else:
        for main_data in mains.values():
            alts = main_data.get("alts", [])
            for alt_candidate in alts:
                if killmail.victim.eve_id == alt_candidate.character_id:
                    main = main_data["main"]
                    alt = alt_candidate
                    break
            if alt:  # Break outer loop if alt is found
                break

    if mode == "attacker":
        main, main_id, alt = killmail.attacker_main(mains)

    return main, main_id, alt


def _stats_killmail(
    killmail: Killmail,
    stats: list,
    mains=None,
    main_list=None,
    title=None,
    count=0,
    mode="killmail",
):
    character_id = killmail.victim.eve_id
    character_name = killmail.victim.name
    portrait = _evulate_portrait(killmail, mode)
    zkb_link = _evaluate_zkb_link(killmail)

    if mains:
        main, _, alt = (
            _get_character_details(killmail, mains, mode)
            if main_list is None
            else main_list
        )
        try:
            character_id = main.character_id if alt is None else alt.character_id
            character_name = (
                f"{main.character_name}"
                if alt is None
                else f"{alt.character_name} ({main.character_name})"
            )
            portrait = eveimageserver.character_portrait_url(character_id, 256)
        except AttributeError:
            logger.debug(
                "Error getting character details for %s %s",
                character_id,
                character_name,
            )

    stats.append(
        {
            # zKB Data
            "killmail_id": killmail.killmail_id,
            "character_id": character_id,
            "character_name": character_name,
            "corporation_id": killmail.victim_corporation_id,
            "alliance_id": killmail.victim_alliance_id,
            "ship": killmail.victim_ship.id,
            "ship_name": killmail.victim_ship.name,
            "hash": killmail.hash,
            "totalValue": killmail.victim_total_value,
            "date": killmail.killmail_date,
            # Additional Data
            "title": f"{killmail.victim.name}" if title is None else title,
            "portrait": portrait,
            "zkb_link": zkb_link,
            "count": count,
        }
    )
    return stats


def killboard_dashboard(
    killmail_year: Killmail,
    date,
    entities,
):
    killmail_year = killmail_year.prefetch_related("victim", "victim_ship")

    stats = []
    stats_manager = KillboardStatsManager(date, stats, entities=entities)
    filtered_killmails_loss = [
        km for km in killmail_year if km.get_month(date.month) and km.is_loss(entities)
    ]
    filtered_killmails_kill = [
        km for km in killmail_year if km.get_month(date.month) and km.is_kill(entities)
    ]

    highest_loss = nlargest(
        1, filtered_killmails_loss, key=lambda km: km.victim_total_value
    )
    highest_kill = nlargest(
        1, filtered_killmails_kill, key=lambda km: km.victim_total_value
    )
    for killmail in killmail_year:
        stats_manager.update_for_killmail(killmail)

    # Update All Stats
    if highest_loss:
        _stats_killmail(highest_loss[0], stats, title="Top Loss:", mode="ship")
    stats_manager.update_worst_ship()
    stats_manager.update_top_victim()
    stats_manager.update_alltime_loss()

    if highest_kill:
        _stats_killmail(highest_kill[0], stats, title="Top Kill:", mode="ship")
    stats_manager.update_top_ship()
    stats_manager.update_top_killer()
    stats_manager.update_alltime_killer()

    return stats


# pylint: disable=too-many-locals
def killboard_hall(killmail_month: Killmail, entities, mains):
    # Ensure that the killmails are sorted by total value
    killmail_month = killmail_month.order_by("-victim_total_value").prefetch_related(
        "victim", "victim_ship"
    )

    shame = []
    fame = []

    topkiller = killmail_month.filter_entities_kills(entities).filter_top_killer(mains)
    toplosses = killmail_month.filter_entities_losses(entities)

    for char in topkiller:
        killmail_fame, alt_char = topkiller.get(char, (None, None))
        if alt_char:
            # Check if the attacker is an alt and get the associated main character
            for main_data in mains.values():
                for alt in main_data["alts"]:
                    if alt_char.character_id == alt.character_id:
                        main = main_data["main"]
                        main_id = char
                        alt_data = alt_char
                        break
        else:
            if char in mains:
                main_data = mains[char]
                main = main_data["main"]
                main_id = char
                alt_data = None

        main_list = [main, main_id, alt_data]

        _stats_killmail(
            killmail_fame, stats=fame, mains=mains, mode="attacker", main_list=main_list
        )

    for killmail in toplosses[:5]:
        _stats_killmail(killmail, stats=shame, mains=mains)

    return shame, fame[:5]



# Standard Library
from typing import Set

# Django
from django.db import models
from django.utils.translation import gettext_lazy as _

# Alliance Auth (External Libs)
from eveuniverse.models import EveType

# Alliance Auth
from allianceauth.eveonline.evelinks import eveimageserver

# AA Killstats
from killstats.hooks import get_extension_logger
from killstats.managers.killboard_manager import EveKillmailManager
from killstats.models.general import EveEntity

logger = get_extension_logger(__name__)


class Killmail(models.Model):
    killmail_id = models.PositiveBigIntegerField(primary_key=True)
    killmail_date = models.DateTimeField(null=True, blank=True)
    victim = models.ForeignKey(EveEntity, on_delete=models.CASCADE, null=True)
    victim_ship = models.ForeignKey(EveType, on_delete=models.CASCADE, null=True)
    victim_corporation_id = models.PositiveBigIntegerField()
    victim_alliance_id = models.PositiveBigIntegerField(null=True, blank=True)
    hash = models.CharField(max_length=255, unique=True)
    # Value Infos
    victim_total_value = models.PositiveBigIntegerField(null=True, blank=True)
    victim_fitted_value = models.PositiveBigIntegerField(null=True, blank=True)
    victim_destroyed_value = models.PositiveBigIntegerField(null=True, blank=True)
    victim_dropped_value = models.PositiveBigIntegerField(null=True, blank=True)
    # Location Infos
    victim_region_id = models.PositiveBigIntegerField(null=True, blank=True)
    victim_solar_system_id = models.PositiveBigIntegerField(null=True, blank=True)
    victim_position_x = models.FloatField(null=True, blank=True)
    victim_position_y = models.FloatField(null=True, blank=True)
    victim_position_z = models.FloatField(null=True, blank=True)
    # Attackers as JSON
    attackers = models.JSONField(null=True, blank=True)

    objects = EveKillmailManager()

    def __str__(self):
        return f"Killmail {self.killmail_id}"

    def get_image_url(self):
        return eveimageserver._eve_entity_image_url(
            self.victim.category, self.victim.eve_id, 32
        )

    def is_kill(self, eve_ids: list):
        """Return True if any attacker is in the list of characters."""
        char_ids = self.attackers_distinct_character_ids()
        corp_ids = self.attackers_distinct_corporation_ids()
        alliance_ids = self.attackers_distinct_alliance_ids()
        if any(char_id in char_ids for char_id in eve_ids):
            return True
        if any(corp_id in corp_ids for corp_id in eve_ids):
            return True
        if any(alliance_id in alliance_ids for alliance_id in eve_ids):
            return True
        return False

    def is_loss(self, eve_ids: list):
        """Return True if the victim is in the list of characters."""
        if self.victim.eve_id in eve_ids:
            return True
        if self.victim_corporation_id in eve_ids:
            return True
        if self.victim_alliance_id in eve_ids:
            return True
        return False

    def is_corp(self, corps: list):
        """Return True if the victim corporation is in the list of corporations."""
        return self.victim_corporation_id in corps or any(
            attacker["corporation_id"] in corps for attacker in self.attackers
        )

    def is_alliance(self, alliances: list):
        """Return True if the victim alliance is in the list of alliances."""
        return self.victim_alliance_id in alliances or any(
            attacker["alliance_id"] in alliances for attacker in self.attackers
        )

    def is_structure(self):
        """Return True if the victim is a structure."""
        return self.victim_ship.eve_group.eve_category_id == 65

    def is_mobile(self):
        """Return True if the victim is a mobile structure."""
        return self.victim_ship.eve_group.eve_category_id == 22

    def is_capsule(self):
        """Return True if the victim is a capsule."""
        return self.victim_ship.id in (670, 33328)

    def get_month(self, month):
        """Get all killmails of a specific month."""
        return self.killmail_date.month == month

    def attackers_distinct_alliance_ids(self) -> Set[int]:
        """Return distinct alliance IDs of all attackers."""
        return [
            attacker["alliance_id"]
            for attacker in self.attackers
            if "alliance_id" in attacker
        ]

    def attackers_distinct_corporation_ids(self) -> Set[int]:
        """Return distinct corporation IDs of all attackers."""
        return [
            attacker["corporation_id"]
            for attacker in self.attackers
            if "corporation_id" in attacker
        ]

    def attackers_distinct_character_ids(self) -> Set[int]:
        """Return distinct character IDs of all attackers."""
        return [
            attacker["character_id"]
            for attacker in self.attackers
            if "character_id" in attacker
        ]

    def threshold(self, threshold: int) -> bool:
        """Return True if the total value of the killmail is above the threshold."""
        return self.victim_total_value > threshold

    def attacker_main(self, mains):
        """
        Get Attacker Main Data from Mains Dict

        Returns
        ----------
        - Main Name:  `str`
        - CharacterID: `int`
        - Alt Object: `object`
        """
        for attacker in self.attackers:
            attacker_id = attacker["character_id"]

            if attacker_id in mains:
                main_data = mains[attacker_id]
                main = main_data["main"]
                main_id = attacker_id
                return main, main_id, None

            # Check if the attacker is an alt and get the associated main character
            for main_data in mains.values():
                for alt in main_data["alts"]:
                    if attacker_id == alt.character_id:
                        main = main_data["main"]
                        main_id = attacker_id
                        attacker = alt
                        return main, main_id, attacker
        return None, None, None

    class Meta:
        default_permissions = ()



from unittest.mock import patch

from django.test import TestCase

from app_utils.esi import EsiDailyDowntime

from killstats.decorators import when_esi_is_available


class TestDecorators(TestCase):
    @patch("killstats.decorators.fetch_esi_status")
    @patch("killstats.decorators.IS_TESTING", new=False)
    def test_when_esi_is_available(self, mock_fetch_esi_status):
        # given
        @when_esi_is_available
        def trigger_esi_deco():
            return "Esi is Available"

        # when
        result = trigger_esi_deco()
        # then
        mock_fetch_esi_status.assert_called_once()
        self.assertEqual(result, "Esi is Available")

    @patch("killstats.decorators.fetch_esi_status", side_effect=EsiDailyDowntime)
    @patch("killstats.decorators.IS_TESTING", new=False)
    def test_when_esi_is_available_downtime(self, mock_fetch_esi_status):
        # given
        @when_esi_is_available
        def trigger_esi_deco():
            return "Daily Downtime detected. Aborting."

        # when
        result = trigger_esi_deco()
        # then
        self.assertIsNone(result)

    @patch("killstats.decorators.fetch_esi_status")
    @patch("killstats.decorators.IS_TESTING", new=True)
    def test_when_esi_is_available_is_test(self, mock_fetch_esi_status):
        # given
        @when_esi_is_available
        def trigger_esi_deco():
            return "Teesting Mode."

        # when
        result = trigger_esi_deco()
        # then
        self.assertEqual(result, "Teesting Mode.")



from allianceauth.eveonline.models import EveCharacter

from killstats.hooks import get_extension_logger

logger = get_extension_logger(__name__)


class KillboardDate:
    def __init__(self, month, year):
        self.month = int(month)
        self.year = int(year)


def get_corporations(request):
    linked_characters = request.user.profile.main_character.character_ownership.user.character_ownerships.select_related(
        "character", "user"
    ).all()

    linked_characters = linked_characters.values_list("character_id", flat=True)
    chars = EveCharacter.objects.filter(id__in=linked_characters)

    corporations = set()

    for char in chars:
        corporations.add(char.corporation_id)

    return list(corporations)


def get_alliances(request):
    linked_characters = request.user.profile.main_character.character_ownership.user.character_ownerships.select_related(
        "character", "user"
    ).all()

    linked_characters = linked_characters.values_list("character_id", flat=True)
    chars = EveCharacter.objects.filter(id__in=linked_characters)

    alliances = set()

    for char in chars:
        alliances.add(char.alliance_id)

    return list(alliances)



from datetime import datetime
from typing import Optional

from ninja import Schema


class Killboard(Schema):
    kills: Optional[int] = None
    date: Optional[datetime] = None
    character_id: Optional[int] = None
    character_name: Optional[str] = None
    ship_id: Optional[int] = None
    ship_name: Optional[str] = None
    hash: Optional[str] = None
    totalValue: Optional[int] = None
    attackers: Optional[list] = None


class KillboardHall(Schema):
    shame: Optional[list] = None
    fame: Optional[list] = None


class KillboardStats(Schema):
    stats: Optional[list] = None


class CorporationAdmin(Schema):
    corporation: Optional[dict] = None


class AllianceAdmin(Schema):
    alliance: Optional[dict] = None



"""API"""

# Standard Library
from typing import List

from ninja import NinjaAPI

# AA Killstats
from killstats.api import schema
from killstats.api.killboard.killboard_helper import (
    get_killmails_data,
    get_killstats_halls,
    get_killstats_stats,
)
from killstats.hooks import get_extension_logger
from killstats.models.killstatsaudit import CorporationsAudit

logger = get_extension_logger(__name__)


class KillboardCorporationApiEndpoints:
    tags = ["Killboard"]

    # pylint: disable=too-many-locals
    def __init__(self, api: NinjaAPI):

        # Killmails
        @api.get(
            "killmail/month/{month}/year/{year}/corporation/{corporation_id}/{mode}/",
            response={200: dict, 403: str},
            tags=self.tags,
        )
        def get_corporation_killmails(
            request, month, year, corporation_id: int, mode, page_size: int = 100
        ):
            return get_killmails_data(
                request, month, year, corporation_id, mode, page_size, "corporation"
            )

        # Hall of Fame/Shame
        @api.get(
            "halls/month/{month}/year/{year}/corporation/{corporation_id}/",
            response={200: List[schema.KillboardHall], 403: str},
            tags=self.tags,
        )
        def get_corporation_halls(request, month, year, corporation_id: int):
            return get_killstats_halls(
                request, month, year, corporation_id, "corporation"
            )

        # Stats
        @api.get(
            "stats/month/{month}/year/{year}/corporation/{corporation_id}/",
            response={200: List[schema.KillboardStats], 403: str},
            tags=self.tags,
        )
        def get_corporation_killstats(request, month, year, corporation_id: int):
            return get_killstats_stats(
                request, month, year, corporation_id, "corporation"
            )

        @api.get(
            "killboard/corporation/admin/",
            response={200: List[schema.CorporationAdmin], 403: str},
            tags=self.tags,
        )
        def get_corporation_admin(request):
            corporations = CorporationsAudit.objects.visible_to(request.user)

            if not corporations:
                return 403, "Permission Denied"

            corporation_dict = {}

            for corporation in corporations:
                # pylint: disable=broad-exception-caught
                try:
                    corporation_dict[corporation.corporation.corporation_id] = {
                        "corporation_id": corporation.corporation.corporation_id,
                        "corporation_name": corporation.corporation.corporation_name,
                    }
                except Exception:
                    continue

            output = []
            output.append({"corporation": corporation_dict})

            return output



from unittest.mock import patch

from ninja import NinjaAPI

from django.test import TestCase

from allianceauth.eveonline.models import EveCorporationInfo
from app_utils.testing import create_user_from_evecharacter

from killstats.api.killboard import KillboardCorporationApiEndpoints
from killstats.tests.test_api import _killstasts_api
from killstats.tests.testdata.load_allianceauth import load_allianceauth
from killstats.tests.testdata.load_killstats import load_killstats_all


class ManageApiJournalCharEndpointsTest(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_allianceauth()
        load_killstats_all()

        cls.user, _ = create_user_from_evecharacter(
            1001,
            permissions=["killstats.basic_access", "killstats.admin_access"],
        )

        cls.user2, _ = create_user_from_evecharacter(
            1002,
            permissions=[
                "killstats.basic_access",
            ],
        )
        cls.api = NinjaAPI()
        cls.manage_api_endpoints = KillboardCorporationApiEndpoints(api=cls.api)

    def test_killstats_killboard_api_no_entry(self):
        self.maxDiff = None
        # given
        self.client.force_login(self.user)
        url = "/killstats/api/stats/month/7/year/2020/corporation/0/"
        # when
        response = self.client.get(url)
        # then
        expected_data = _killstasts_api.Killstats_Stats
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json(), expected_data)

    def test_killstats_killboard_api(self):
        self.maxDiff = None
        # given
        self.client.force_login(self.user)
        url = "/killstats/api/stats/month/7/year/2024/corporation/0/"
        # when
        response = self.client.get(url)
        # then
        expected_data = _killstasts_api.Killstats_Stats_Entry
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json(), expected_data)

    def test_killstats_killboard_api_single(self):
        self.maxDiff = None
        # given
        self.client.force_login(self.user)
        url = "/killstats/api/stats/month/7/year/2024/corporation/2001/"
        # when
        response = self.client.get(url)
        # then
        expected_data = _killstasts_api.Killstats_Stats_Entry
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json(), expected_data)

        url = "/killstats/api/halls/month/7/year/2024/corporation/2001/"
        # when
        response = self.client.get(url)
        # then
        expected_data = _killstasts_api.Killstats_Halls_Entry
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json(), expected_data)

        url = "/killstats/api/killmail/month/7/year/2024/corporation/2001/kills/"
        # when
        response = self.client.get(url)

        # then
        expected_data = _killstasts_api.Killstats_Kills_Entry
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json(), expected_data)

    def test_get_corporation_admin(self):
        self.client.force_login(self.user2)
        url = "/killstats/api/killboard/corporation/admin/"
        # when
        response = self.client.get(url)
        # then
        excepted_data = [
            {
                "corporation": {
                    "2002": {"corporation_id": 2002, "corporation_name": "Eulenclub"}
                }
            }
        ]
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json(), excepted_data)

    @patch(
        "killstats.api.killboard.corporation.killboard.CorporationsAudit.objects.visible_to"
    )
    def test_get_corporation_admin_no_visible(self, mock_visible_to):
        self.client.force_login(self.user2)
        url = "/killstats/api/killboard/corporation/admin/"

        mock_visible_to.return_value = None

        # when
        response = self.client.get(url)
        # then
        self.assertContains(response, "Permission Denied", status_code=403)

    @patch(
        "killstats.api.killboard.corporation.killboard.CorporationsAudit.objects.visible_to"
    )
    def test_get_corporation_admin_exception(self, mock_visible_to):
        self.client.force_login(self.user)
        url = "/killstats/api/killboard/corporation/admin/"

        corp = EveCorporationInfo.objects.get(corporation_id=2001)

        mock_visible_to.return_value = [corp, "test"]

        # when
        response = self.client.get(url)
        # then
        self.assertEqual(response.status_code, 200)



# Generated by Django 4.2.11 on 2024-08-15 11:51

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("eveonline", "0017_alliance_and_corp_names_are_not_unique"),
        ("killstats", "0001_initial"),
    ]

    operations = [
        migrations.RenameModel(
            old_name="KillstatsAudit",
            new_name="CorporationsAudit",
        ),
        migrations.CreateModel(
            name="AlliancesAudit",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("last_update", models.DateTimeField(auto_now=True)),
                (
                    "alliance",
                    models.OneToOneField(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="eveonline.eveallianceinfo",
                    ),
                ),
                (
                    "owner",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="eveonline.evecharacter",
                    ),
                ),
            ],
            options={
                "default_permissions": (),
            },
        ),
    ]



from django.test import TestCase

from killstats.models.killstatsaudit import AlliancesAudit, CorporationsAudit
from killstats.tests.testdata.load_allianceauth import load_allianceauth
from killstats.tests.testdata.load_killstats import load_killstats_all

MODULE_PATH = "killstats.models.killstatsaudit"


class TestKillstatsAuditModel(TestCase):
    @classmethod
    def setUp(self):
        load_allianceauth()
        load_killstats_all()

    def test_str(self):
        self.audit = CorporationsAudit.objects.get(corporation__corporation_id=2001)
        self.assertEqual(str(self.audit), "Hell RiderZ's Killstats Data")

    def test_alliance_str(self):
        self.audit = AlliancesAudit.objects.get(alliance__alliance_id=3002)
        self.assertEqual(str(self.audit), "Eulen Sigma's Killstats Data")



from .alliance.killboard import KillboardAllianceApiEndpoints
from .corporation.killboard import KillboardCorporationApiEndpoints


def setup(api):
    KillboardCorporationApiEndpoints(api)
    KillboardAllianceApiEndpoints(api)



KillstatsMonth = [
    {
        "kills": [],
        "losses": [
            {
                "title": "None",
                "killmail_id": 119303113,
                "name": "Gneuten",
                "character_id": 1001,
                "character_name": "Gneuten",
                "portrait": "https://images.evetech.net/types/17634/icon?size=64",
                "corporation_id": 2001,
                "alliance_id": 3001,
                "ship": 17634,
                "ship_name": "Caracal Navy Issue",
                "hash": "95ff494261eed2c4a790aad70c7f0053c75f89c6",
                "totalValue": 76734851,
                "date": "2024-07-07T21:38:59Z",
                "count": 0,
            },
            {
                "title": "None",
                "killmail_id": 119303075,
                "name": "Garthd",
                "character_id": 1004,
                "character_name": "Garthd",
                "portrait": "https://images.evetech.net/types/29340/icon?size=64",
                "corporation_id": 2001,
                "alliance_id": 3001,
                "ship": 29340,
                "ship_name": "Osprey Navy Issue",
                "hash": "1185d04b1b6c6667167aa7d65245a49d19d44275",
                "totalValue": 1000,
                "date": "2024-07-07T21:38:58Z",
                "count": 0,
            },
            {
                "title": "None",
                "killmail_id": 119303073,
                "name": "rotze Rotineque",
                "character_id": 1002,
                "character_name": "rotze Rotineque",
                "portrait": "https://images.evetech.net/types/29340/icon?size=64",
                "corporation_id": 2001,
                "alliance_id": 3001,
                "ship": 29340,
                "ship_name": "Osprey Navy Issue",
                "hash": "1185d04b1b6c5467167aa7d65245a49d19d44275",
                "totalValue": 94395699,
                "date": "2024-07-07T21:36:58Z",
                "count": 0,
            },
        ],
        "totalKills": 0,
        "totalLoss": 171131550,
        "shame": [
            {
                "title": "None",
                "killmail_id": 119303073,
                "name": "rotze Rotineque",
                "character_id": 1002,
                "character_name": "rotze Rotineque",
                "portrait": "https://images.evetech.net/types/29340/icon?size=64",
                "corporation_id": 2001,
                "alliance_id": 3001,
                "ship": 29340,
                "ship_name": "Osprey Navy Issue",
                "hash": "1185d04b1b6c5467167aa7d65245a49d19d44275",
                "totalValue": 94395699,
                "date": "2024-07-07T21:36:58Z",
                "count": 0,
            },
            {
                "title": "None",
                "killmail_id": 119303113,
                "name": "Gneuten",
                "character_id": 1001,
                "character_name": "Gneuten",
                "portrait": "https://images.evetech.net/types/17634/icon?size=64",
                "corporation_id": 2001,
                "alliance_id": 3001,
                "ship": 17634,
                "ship_name": "Caracal Navy Issue",
                "hash": "95ff494261eed2c4a790aad70c7f0053c75f89c6",
                "totalValue": 76734851,
                "date": "2024-07-07T21:38:59Z",
                "count": 0,
            },
            {
                "title": "None",
                "killmail_id": 119303075,
                "name": "Garthd",
                "character_id": 1004,
                "character_name": "Garthd",
                "portrait": "https://images.evetech.net/types/29340/icon?size=64",
                "corporation_id": 2001,
                "alliance_id": 3001,
                "ship": 29340,
                "ship_name": "Osprey Navy Issue",
                "hash": "1185d04b1b6c6667167aa7d65245a49d19d44275",
                "totalValue": 1000,
                "date": "2024-07-07T21:38:58Z",
                "count": 0,
            },
        ],
        "fame": [],
        "stats": [
            {
                "title": "Top Loss:",
                "killmail_id": 119303073,
                "name": "rotze Rotineque",
                "character_id": 1002,
                "character_name": "rotze Rotineque",
                "portrait": "https://images.evetech.net/types/29340/icon?size=64",
                "corporation_id": 2001,
                "alliance_id": 3001,
                "ship": 29340,
                "ship_name": "Osprey Navy Issue",
                "hash": "1185d04b1b6c5467167aa7d65245a49d19d44275",
                "totalValue": 94395699,
                "date": "2024-07-07T21:36:58Z",
                "count": 0,
            },
            {
                "title": "Worst Ship:",
                "type": "count",
                "ship": 29340,
                "ship_name": "Osprey Navy Issue",
                "portrait": "https://images.evetech.net/types/29340/icon?size=256",
                "count": 2,
                "loss": True,
            },
            {
                "title": "Top Victim: Gneuten",
                "type": "count",
                "character_id": 1001,
                "character_name": "Gneuten",
                "portrait": "https://images.evetech.net/characters/1001/portrait?size=256",
                "count": 1,
                "loss": True,
            },
            {
                "title": "Alltime Victim: Gneuten",
                "type": "count",
                "character_id": 1001,
                "character_name": "Gneuten",
                "portrait": "https://images.evetech.net/characters/1001/portrait?size=256",
                "count": 1,
                "loss": True,
            },
            {
                "title": "Top Ship:",
                "type": "count",
                "ship": 17634,
                "ship_name": "Caracal Navy Issue",
                "portrait": "https://images.evetech.net/types/17634/icon?size=256",
                "count": 1,
                "loss": False,
            },
            {
                "title": "Top Killer: Gneuten",
                "type": "count",
                "character_id": 1001,
                "character_name": "Gneuten",
                "portrait": "https://images.evetech.net/characters/1001/portrait?size=256",
                "count": 1,
                "loss": False,
            },
            {
                "title": "Alltime Killer: Gneuten",
                "type": "count",
                "character_id": 1001,
                "character_name": "Gneuten",
                "portrait": "https://images.evetech.net/characters/1001/portrait?size=256",
                "count": 1,
                "loss": False,
            },
        ],
    }
]

KillstatsEmpty = [
    {
        "kills": [],
        "losses": [],
        "totalKills": 0,
        "totalLoss": 0,
        "shame": [],
        "fame": [],
        "stats": [],
    }
]

Killstats_Stats = [{"stats": []}]

Killstats_Stats_Entry = [
    {
        "stats": [
            {
                "killmail_id": 119303073,
                "character_id": 1002,
                "character_name": "rotze Rotineque",
                "corporation_id": 2001,
                "alliance_id": 3001,
                "ship": 29340,
                "ship_name": "Osprey Navy Issue",
                "hash": "1185d04b1b6c5467167aa7d65245a49d19d44275",
                "totalValue": 94395699,
                "date": "2024-07-07T21:36:58Z",
                "title": "Top Loss:",
                "portrait": "https://images.evetech.net/types/29340/icon?size=256",
                "zkb_link": "https://zkillboard.com/character/1002/",
                "count": 0,
            },
            {
                "title": "Worst Ship:",
                "type": "count",
                "ship": 29340,
                "ship_name": "Osprey Navy Issue",
                "portrait": "https://images.evetech.net/types/29340/icon?size=256",
                "count": 2,
                "loss": True,
                "zkb_link": "https://zkillboard.com/character/29340/",
            },
            {
                "title": "Top Victim: Gneuten",
                "type": "count",
                "character_id": 1001,
                "character_name": "Gneuten",
                "portrait": "https://images.evetech.net/characters/1001/portrait?size=256",
                "count": 1,
                "loss": True,
                "zkb_link": "https://zkillboard.com/character/1001/",
            },
            {
                "title": "Alltime Victim: Gneuten",
                "type": "count",
                "character_id": 1001,
                "character_name": "Gneuten",
                "portrait": "https://images.evetech.net/characters/1001/portrait?size=256",
                "count": 1,
                "loss": True,
                "zkb_link": "https://zkillboard.com/character/1001/",
            },
            {
                "killmail_id": 119303075,
                "character_id": 1004,
                "character_name": "Garthd",
                "corporation_id": 2001,
                "alliance_id": 3001,
                "ship": 29340,
                "ship_name": "Osprey Navy Issue",
                "hash": "1185d04b1b6c6667167aa7d65245a49d19d44275",
                "totalValue": 1000,
                "date": "2024-07-07T21:38:58Z",
                "title": "Top Kill:",
                "portrait": "https://images.evetech.net/types/29340/icon?size=256",
                "zkb_link": "https://zkillboard.com/character/1004/",
                "count": 0,
            },
            {
                "title": "Top Ship:",
                "type": "count",
                "ship": 17634,
                "ship_name": "Caracal Navy Issue",
                "portrait": "https://images.evetech.net/types/17634/icon?size=256",
                "count": 1,
                "loss": False,
                "zkb_link": "https://zkillboard.com/character/17634/",
            },
            {
                "title": "Top Killer: Gneuten",
                "type": "count",
                "character_id": 1001,
                "character_name": "Gneuten",
                "portrait": "https://images.evetech.net/characters/1001/portrait?size=256",
                "count": 1,
                "loss": False,
                "zkb_link": "https://zkillboard.com/character/1001/",
            },
            {
                "title": "Alltime Killer: Gneuten",
                "type": "count",
                "character_id": 1001,
                "character_name": "Gneuten",
                "portrait": "https://images.evetech.net/characters/1001/portrait?size=256",
                "count": 1,
                "loss": False,
                "zkb_link": "https://zkillboard.com/character/1001/",
            },
        ]
    }
]


Killstats_Halls_Entry = [
    {
        "shame": [
            {
                "killmail_id": 119303073,
                "character_id": 1002,
                "character_name": "rotze Rotineque",
                "corporation_id": 2001,
                "alliance_id": 3001,
                "ship": 29340,
                "ship_name": "Osprey Navy Issue",
                "hash": "1185d04b1b6c5467167aa7d65245a49d19d44275",
                "totalValue": 94395699,
                "date": "2024-07-07T21:36:58Z",
                "title": "rotze Rotineque",
                "portrait": "https://images.evetech.net/characters/1002/portrait?size=256",
                "zkb_link": "https://zkillboard.com/character/1002/",
                "count": 0,
            },
            {
                "killmail_id": 119303113,
                "character_id": 1001,
                "character_name": "Gneuten",
                "corporation_id": 2001,
                "alliance_id": 3001,
                "ship": 17634,
                "ship_name": "Caracal Navy Issue",
                "hash": "95ff494261eed2c4a790aad70c7f0053c75f89c6",
                "totalValue": 76734851,
                "date": "2024-07-07T21:38:59Z",
                "title": "Gneuten",
                "portrait": "https://images.evetech.net/characters/1001/portrait?size=256",
                "zkb_link": "https://zkillboard.com/character/1001/",
                "count": 0,
            },
            {
                "killmail_id": 119303075,
                "character_id": 1004,
                "character_name": "Garthd",
                "corporation_id": 2001,
                "alliance_id": 3001,
                "ship": 29340,
                "ship_name": "Osprey Navy Issue",
                "hash": "1185d04b1b6c6667167aa7d65245a49d19d44275",
                "totalValue": 1000,
                "date": "2024-07-07T21:38:58Z",
                "title": "Garthd",
                "portrait": "https://images.evetech.net/characters/1004/portrait?size=256",
                "zkb_link": "https://zkillboard.com/character/1004/",
                "count": 0,
            },
        ],
        "fame": [
            {
                "killmail_id": 119303075,
                "character_id": 1001,
                "character_name": "Gneuten",
                "corporation_id": 2001,
                "alliance_id": 3001,
                "ship": 29340,
                "ship_name": "Osprey Navy Issue",
                "hash": "1185d04b1b6c6667167aa7d65245a49d19d44275",
                "totalValue": 1000,
                "date": "2024-07-07T21:38:58Z",
                "title": "Garthd",
                "portrait": "https://images.evetech.net/characters/1001/portrait?size=256",
                "zkb_link": "https://zkillboard.com/character/1004/",
                "count": 0,
            },
            {
                "killmail_id": 119303075,
                "character_id": 1002,
                "character_name": "rotze Rotineque",
                "corporation_id": 2001,
                "alliance_id": 3001,
                "ship": 29340,
                "ship_name": "Osprey Navy Issue",
                "hash": "1185d04b1b6c6667167aa7d65245a49d19d44275",
                "totalValue": 1000,
                "date": "2024-07-07T21:38:58Z",
                "title": "Garthd",
                "portrait": "https://images.evetech.net/characters/1002/portrait?size=256",
                "zkb_link": "https://zkillboard.com/character/1004/",
                "count": 0,
            },
        ],
    }
]

Killstats_Kills_Entry = {
    "draw": 1,
    "recordsTotal": 1,
    "recordsFiltered": 1,
    "data": [
        {
            "killmail_id": 119303075,
            "killmail_date": "2024-07-07T21:38:58Z",
            "victim": {"id": 1004, "name": "Garthd"},
            "victim_ship": {"id": 29340, "name": "Osprey Navy Issue"},
            "victim_corporation_id": 2001,
            "victim_alliance_id": 3001,
            "hash": "1185d04b1b6c6667167aa7d65245a49d19d44275",
            "victim_total_value": 1000,
            "victim_fitted_value": 2000,
            "victim_destroyed_value": 3000,
            "victim_dropped_value": 1000,
            "victim_region_id": 10000059,
            "victim_solar_system_id": 30004668,
            "victim_position_x": -74459258618.29333,
            "victim_position_y": -4921962525.529236,
            "victim_position_z": -8888733393.073042,
            "attackers": [
                {
                    "character_id": 1001,
                    "corporation_id": 2001,
                    "alliance_id": 3001,
                    "ship_type_id": 17634,
                    "damage_done": 8056,
                    "final_blow": False,
                },
                {
                    "character_id": 1002,
                    "corporation_id": 2001,
                    "alliance_id": 3001,
                    "ship_type_id": 29340,
                    "damage_done": 2788,
                    "final_blow": True,
                },
            ],
        }
    ],
    "totalvalue": [1000],
}



"""Shared ESI client for Voices of War."""

# Alliance Auth
from esi.clients import EsiClientProvider

# AA Killstats
from killstats import __version__

esi = EsiClientProvider(app_info_text=f"killstats v{__version__}")



from unittest.mock import MagicMock, patch

from django.contrib.admin.sites import AdminSite
from django.contrib.messages import get_messages
from django.contrib.messages.middleware import MessageMiddleware
from django.contrib.sessions.middleware import SessionMiddleware
from django.test import RequestFactory, TestCase

from allianceauth.eveonline.evelinks import eveimageserver
from app_utils.testdata_factories import UserFactory

from killstats.admin import (
    AlliancesAuditAdmin,
    CorporationsAuditAdmin,
    clear_alliance_cache_zkb,
    clear_cache_zkb,
)
from killstats.models.killstatsaudit import AlliancesAudit, CorporationsAudit
from killstats.tests.testdata.load_allianceauth import load_allianceauth
from killstats.tests.testdata.load_killstats import load_killstats_all

MODULE_PATH = "killstats.admin"


class MockRequest:
    pass


class TestKillstatsAuditAdmin(TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        load_allianceauth()
        load_killstats_all()
        cls.factory = RequestFactory()
        cls.site = AdminSite()
        cls.killstats_audit_admin = CorporationsAuditAdmin(CorporationsAudit, cls.site)
        cls.killstats_audit = CorporationsAudit.objects.get(id=1)

    def test_entity_pic(self):
        user = UserFactory(is_superuser=True, is_staff=True)
        self.client.force_login(user)
        request = self.factory.get("/")
        request.user = user
        expected_html = '<img src="{}" class="img-circle">'.format(
            eveimageserver._eve_entity_image_url(
                "corporation", self.killstats_audit.corporation.corporation_id, 32
            )
        )
        self.assertEqual(
            self.killstats_audit_admin._entity_pic(self.killstats_audit),
            expected_html,
        )

    def test_corporation_corporation_id(self):
        user = UserFactory(is_superuser=True, is_staff=True)
        self.client.force_login(user)
        request = self.factory.get("/")
        request.user = user
        self.assertEqual(
            self.killstats_audit_admin._corporation__corporation_id(
                self.killstats_audit
            ),
            2001,
        )

    def test_last_update(self):
        user = UserFactory(is_superuser=True, is_staff=True)
        self.client.force_login(user)
        request = self.factory.get("/")
        request.user = user
        self.assertEqual(
            self.killstats_audit_admin._last_update(self.killstats_audit),
            self.killstats_audit.last_update,
        )

    def test_has_add_permission(self):
        user = UserFactory(is_superuser=True, is_staff=True)
        self.client.force_login(user)
        request = self.factory.get("/")
        request.user = user
        self.assertFalse(self.killstats_audit_admin.has_add_permission(request))

    def test_has_change_permission(self):
        user = UserFactory(is_superuser=True, is_staff=True)
        self.client.force_login(user)
        request = self.factory.get("/")
        request.user = user
        request = MockRequest()
        self.assertFalse(self.killstats_audit_admin.has_change_permission(request))
        self.assertFalse(
            self.killstats_audit_admin.has_change_permission(
                request, obj=self.killstats_audit
            )
        )

    @patch("killstats.admin.clear_cache_zkb")
    def test_clear_cache_for_selected(self, mock_clear_cache_zkb):
        user = UserFactory(is_superuser=True, is_staff=True)
        self.client.force_login(user)
        request = self.factory.post("/")
        request.user = user

        # Fügen Sie die MessageMiddleware hinzu
        session_middleware = SessionMiddleware(lambda req: None)
        message_middleware = MessageMiddleware(lambda req: None)
        session_middleware.process_request(request)
        request.session = self.client.session
        message_middleware.process_request(request)

        queryset = CorporationsAudit.objects.filter(id=self.killstats_audit.pk)

        self.killstats_audit_admin.clear_cache_for_selected(request, queryset)

        mock_clear_cache_zkb.assert_called_once_with(
            self.killstats_audit.corporation.corporation_id
        )

        # Überprüfen, dass die Nachricht an den Benutzer gesendet wurde
        messages = list(get_messages(request))
        self.assertEqual(len(messages), 1)
        self.assertEqual(
            str(messages[0]), "Cache successfully cleared for selected Corporations."
        )

    @patch(MODULE_PATH + ".get_redis_connection")
    def test_clear_cache_zkb(self, mock_get_redis_connection):
        # Mock the Redis connection
        mock_conn = MagicMock()
        mock_get_redis_connection.return_value = mock_conn

        # Mock the scan_iter method to return a list of keys
        mock_conn.scan_iter.return_value = [
            "zkb_page_cache_https://zkillboard.com/api/npc/0/corporationID/12345_1",
            "zkb_page_cache_https://zkillboard.com/api/npc/0/corporationID/12345_2",
            "zkb_page_cache_https://zkillboard.com/api/npc/0/corporationID/12345_3",
        ]

        # Call the function with a test corporation_id
        result = clear_cache_zkb(12345)

        # Verify that the delete method was called with the correct keys
        mock_conn.delete.assert_any_call(
            "zkb_page_cache_https://zkillboard.com/api/npc/0/corporationID/12345_1"
        )
        mock_conn.delete.assert_any_call(
            "zkb_page_cache_https://zkillboard.com/api/npc/0/corporationID/12345_2"
        )
        mock_conn.delete.assert_any_call(
            "zkb_page_cache_https://zkillboard.com/api/npc/0/corporationID/12345_3"
        )

        # Verify that the function returns True
        self.assertTrue(result)

        # Verify that scan_iter was called with the correct pattern
        mock_conn.scan_iter.assert_called_once_with(
            "*:zkb_page_cache_https://zkillboard.com/api/npc/0/corporationID/12345*"
        )


class TestAlliancesAuditAdmin(TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        load_allianceauth()
        load_killstats_all()
        cls.factory = RequestFactory()
        cls.site = AdminSite()
        cls.killstats_audit_admin = AlliancesAuditAdmin(AlliancesAudit, cls.site)
        cls.killstats_audit = AlliancesAudit.objects.get(id=1)

    def test_entity_pic(self):
        user = UserFactory(is_superuser=True, is_staff=True)
        self.client.force_login(user)
        request = self.factory.get("/")
        request.user = user
        expected_html = '<img src="{}" class="img-circle">'.format(
            eveimageserver._eve_entity_image_url(
                "alliance", self.killstats_audit.alliance.alliance_id, 32
            )
        )
        self.assertEqual(
            self.killstats_audit_admin._entity_pic(self.killstats_audit),
            expected_html,
        )

    def test_alliance_alliance_id(self):
        user = UserFactory(is_superuser=True, is_staff=True)
        self.client.force_login(user)
        request = self.factory.get("/")
        request.user = user
        self.assertEqual(
            self.killstats_audit_admin._alliance__alliance_id(self.killstats_audit),
            3001,
        )

    def test_last_update(self):
        user = UserFactory(is_superuser=True, is_staff=True)
        self.client.force_login(user)
        request = self.factory.get("/")
        request.user = user
        self.assertEqual(
            self.killstats_audit_admin._last_update(self.killstats_audit),
            self.killstats_audit.last_update,
        )

    def test_has_add_permission(self):
        user = UserFactory(is_superuser=True, is_staff=True)
        self.client.force_login(user)
        request = self.factory.get("/")
        request.user = user
        self.assertFalse(self.killstats_audit_admin.has_add_permission(request))

    def test_has_change_permission(self):
        user = UserFactory(is_superuser=True, is_staff=True)
        self.client.force_login(user)
        request = self.factory.get("/")
        request.user = user
        request = MockRequest()
        self.assertFalse(self.killstats_audit_admin.has_change_permission(request))
        self.assertFalse(
            self.killstats_audit_admin.has_change_permission(
                request, obj=self.killstats_audit
            )
        )

    @patch("killstats.admin.clear_alliance_cache_zkb")
    def test_clear_alliance_cache_for_selected(self, mock_clear_cache_zkb):
        user = UserFactory(is_superuser=True, is_staff=True)
        self.client.force_login(user)
        request = self.factory.post("/")
        request.user = user

        # Fügen Sie die MessageMiddleware hinzu
        session_middleware = SessionMiddleware(lambda req: None)
        message_middleware = MessageMiddleware(lambda req: None)
        session_middleware.process_request(request)
        request.session = self.client.session
        message_middleware.process_request(request)

        queryset = AlliancesAudit.objects.filter(id=self.killstats_audit.pk)

        self.killstats_audit_admin.clear_cache_for_selected(request, queryset)

        mock_clear_cache_zkb.assert_called_once_with(
            self.killstats_audit.alliance.alliance_id
        )

        # Überprüfen, dass die Nachricht an den Benutzer gesendet wurde
        messages = list(get_messages(request))
        self.assertEqual(len(messages), 1)
        self.assertEqual(
            str(messages[0]), "Cache successfully cleared for selected Alliances."
        )

    @patch(MODULE_PATH + ".get_redis_connection")
    def test_clear_alliance_cache_zkb(self, mock_get_redis_connection):
        # Mock the Redis connection
        mock_conn = MagicMock()
        mock_get_redis_connection.return_value = mock_conn

        # Mock the scan_iter method to return a list of keys
        mock_conn.scan_iter.return_value = [
            "zkb_page_cache_https://zkillboard.com/api/npc/0/allianceID/12345_1",
            "zkb_page_cache_https://zkillboard.com/api/npc/0/allianceID/12345_2",
            "zkb_page_cache_https://zkillboard.com/api/npc/0/allianceID/12345_3",
        ]

        # Call the function with a test alliance_id
        result = clear_alliance_cache_zkb(12345)

        # Verify that the delete method was called with the correct keys
        mock_conn.delete.assert_any_call(
            "zkb_page_cache_https://zkillboard.com/api/npc/0/allianceID/12345_1"
        )
        mock_conn.delete.assert_any_call(
            "zkb_page_cache_https://zkillboard.com/api/npc/0/allianceID/12345_2"
        )
        mock_conn.delete.assert_any_call(
            "zkb_page_cache_https://zkillboard.com/api/npc/0/allianceID/12345_3"
        )

        # Verify that the function returns True
        self.assertTrue(result)

        # Verify that scan_iter was called with the correct pattern
        mock_conn.scan_iter.assert_called_once_with(
            "*:zkb_page_cache_https://zkillboard.com/api/npc/0/allianceID/12345*"
        )



"""Managers for killboard."""

# Standard Library
from typing import Any, Tuple

# Django
from django.db import models, transaction

# AA Voices of War
from killstats.hooks import get_extension_logger
from killstats.managers.killmail_core import KillmailManager

logger = get_extension_logger(__name__)


class KillmailQueryCore(models.QuerySet):
    def filter_entities(self, entities):
        """Filter Kills and Losses from Entities List (Corporations or Alliances)."""
        kms = []
        for killmail in self:
            if any(
                attacker["corporation_id"] in entities
                for attacker in killmail.attackers
            ):
                kms.append(killmail.killmail_id)
            if killmail.victim_corporation_id in entities:
                kms.append(killmail.killmail_id)
            if any(
                attacker["alliance_id"] in entities for attacker in killmail.attackers
            ):
                kms.append(killmail.killmail_id)
            if killmail.victim_alliance_id in entities:
                kms.append(killmail.killmail_id)
        return self.filter(killmail_id__in=kms)

    def filter_entities_kills(self, entities):
        """Filter Kills from Entities List (Corporations or Alliances)."""
        kms = []
        for killmail in self:
            if any(
                attacker["corporation_id"] in entities
                for attacker in killmail.attackers
            ):
                kms.append(killmail.killmail_id)
            if any(
                attacker["alliance_id"] in entities for attacker in killmail.attackers
            ):
                kms.append(killmail.killmail_id)
        return self.filter(killmail_id__in=kms)

    def filter_entities_losses(self, entities):
        """Filter Losses from Entities List (Corporations or Alliances)."""
        kms = []
        for killmail in self:
            if killmail.victim.eve_id in entities:
                kms.append(killmail.killmail_id)
            if killmail.victim_corporation_id in entities:
                kms.append(killmail.killmail_id)
            if killmail.victim_alliance_id in entities:
                kms.append(killmail.killmail_id)
        return self.filter(killmail_id__in=kms)

    def filter_structure(self, exclude=False):
        """Filter or Exclude Structure Kills."""
        if exclude:
            return self.exclude(victim_ship__eve_group__eve_category_id=65)
        return self.filter(victim_ship__eve_group__eve_category_id=65)

    def filter_threshold(self, threshold: int):
        """Filter Killmails are in Threshold."""
        return self.filter(victim_total_value__gt=threshold)

    def filter_top_killer(self, mains):
        """Returns Topkiller from Killmail as dict."""
        topkiller = {}
        for killmail in self:
            for attacker in killmail.attackers:
                character_id, alt_char = self._find_main_or_alt(attacker, mains)
                if not character_id:
                    continue
                self._update_topkiller(topkiller, character_id, killmail, alt_char)
        return topkiller

    def _find_main_or_alt(self, attacker, mains):
        """Finds the main character or alt character based on attacker info."""
        character_id = attacker.get("character_id", 0)
        if character_id in mains:
            main_data = mains[character_id]
            return main_data["main"].character_id, None
        for main_data in mains.values():
            for alt in main_data["alts"]:
                if character_id == alt.character_id:
                    return main_data["main"].character_id, alt
        return None, None

    def _update_topkiller(self, topkiller, character_id, killmail, alt_char):
        """Updates the topkiller dictionary with the highest value killmail."""
        current_highest_value, _ = topkiller.get(character_id, (None, None))
        if (
            current_highest_value is None
            or killmail.victim_total_value > current_highest_value.victim_total_value
        ):
            topkiller[character_id] = (killmail, alt_char)


class KillmailQueryMining(KillmailQueryCore):
    def filter_barge(self):
        """Filter Mining Barge."""
        return self.filter(victim_ship__eve_group_id=463)

    def filter_exhumer(self):
        """Filter Exhumer."""
        return self.filter(victim_ship__eve_group_id=543)

    def filter_indu_command_ship(self):
        """Filter Industrial Command Ship."""
        return self.filter(victim_ship__eve_group_id=941)

    def filter_capital_indu_ship(self):
        """Filter Capital Industrial Ship."""
        return self.filter(victim_ship__eve_group_id=883)


class KillmailQuerySet(KillmailQueryMining):
    def visible_to(self, user):
        # superusers get all visible
        if user.is_superuser:
            logger.debug("Returning all Squads for superuser %s.", user)
            return self

        if user.has_perm("killstats.admin_access"):
            logger.debug("Returning all Killboards for Admin %s.", user)
            return self

        return self.none()


class KillmailBaseManager(models.Manager):
    def get_queryset(self):
        return KillmailQuerySet(self.model, using=self._db)

    def visible_to(self, user):
        return self.get_queryset().visible_to(user)

    def create_from_killmail(self, killmail: KillmailManager):
        """create a new EveKillmail from a Killmail object and returns it"""
        # AA Killstats
        # pylint: disable=import-outside-toplevel
        from killstats.models.killboard import Killmail

        attackers_list = []

        corporation_id = killmail.victim.corporation_id
        region_id = killmail.get_region_id(killmail.solar_system_id)
        victim_ship = killmail.get_ship_name(killmail.victim.ship_type_id)
        victim = None

        if killmail.victim.character_id:
            victim = killmail.get_entity_name(killmail.victim.character_id)
        elif killmail.victim.alliance_id:
            victim = killmail.get_entity_name(killmail.victim.alliance_id)
        elif killmail.victim.corporation_id:
            victim = killmail.get_entity_name(killmail.victim.corporation_id)

        for attacker in killmail.attackers:
            if attacker.character_id:
                attackers_list.append(attacker.character_id)

        if attackers_list:
            try:
                unique_attackers = list(set(attackers_list))
                killmail.create_names_bulk(eve_ids=unique_attackers)
            # pylint: disable=broad-exception-caught
            except Exception as e:
                logger.debug("Error on Create Names: %s", e, exc_info=True)

        _, new_entry = Killmail.objects.get_or_create(
            killmail_id=killmail.id,
            killmail_date=killmail.time,
            victim=victim,
            victim_ship=victim_ship,
            victim_corporation_id=corporation_id,
            victim_alliance_id=killmail.victim.alliance_id,
            hash=killmail.zkb.hash,
            victim_total_value=killmail.zkb.total_value,
            victim_fitted_value=killmail.zkb.fitted_value,
            victim_destroyed_value=killmail.zkb.destroyed_value,
            victim_dropped_value=killmail.zkb.dropped_value,
            victim_region_id=region_id,
            victim_solar_system_id=killmail.solar_system_id,
            victim_position_x=killmail.position.x,
            victim_position_y=killmail.position.y,
            victim_position_z=killmail.position.z,
            attackers=killmail.attackers_distinct_info(),
        )
        return new_entry

    def update_or_create_from_killmail(
        self, killmail: KillmailManager
    ) -> Tuple[Any, bool]:
        """Update or create new EveKillmail from a Killmail object."""
        with transaction.atomic():
            try:
                self.get(id=killmail.id).delete()
                created = False
            except self.model.DoesNotExist:
                created = True
            obj = self.create_from_killmail(killmail)
        return obj, created


EveKillmailManager = KillmailBaseManager.from_queryset(KillmailQuerySet)



from datetime import datetime

from django.test import TestCase

from allianceauth.eveonline.evelinks import eveimageserver
from allianceauth.eveonline.models import EveCharacter

from killstats.models.killboard import EveEntity, EveType, Killmail
from killstats.tests.testdata.load_allianceauth import load_allianceauth
from killstats.tests.testdata.load_killstats import load_killstats_all

MODULE_PATH = "killstats.models.killstatsaudit"


class TestKillboardtModel(TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        load_allianceauth()
        load_killstats_all()
        cls.victim = EveEntity.objects.get(eve_id=1005)
        cls.victim_ship = EveType.objects.get(id=670)
        cls.killmail = Killmail.objects.create(
            killmail_id=1,
            killmail_date=datetime.now(),
            victim=cls.victim,
            victim_ship=cls.victim_ship,
            victim_corporation_id=2001,
            hash="hash1",
            attackers=[
                {"character_id": 1001, "corporation_id": 2001, "alliance_id": 3001},
                {"character_id": 1002, "corporation_id": 2002, "alliance_id": 3002},
            ],
            victim_total_value=1_000_000,
        )

        cls.killmail2 = Killmail.objects.create(
            killmail_id=2,
            killmail_date=datetime.now(),
            victim=cls.victim,
            victim_ship=cls.victim_ship,
            victim_corporation_id=2001,
            hash="hash2",
            attackers=[
                {"character_id": 1004, "corporation_id": 2001, "alliance_id": 3001},
                {"character_id": 1005, "corporation_id": 2002, "alliance_id": 3002},
            ],
            victim_total_value=100_000_000,
        )

    def test_is_kill(self):
        self.assertTrue(self.killmail.is_kill([1002]))
        self.assertFalse(self.killmail.is_kill([1004]))

    def test_is_loss(self):
        self.assertTrue(self.killmail.is_loss([1005]))
        self.assertFalse(self.killmail.is_loss([1004]))

    def test_is_corp(self):
        self.assertTrue(self.killmail.is_corp([2001]))
        self.assertTrue(self.killmail.is_corp([2001]))
        self.assertFalse(self.killmail.is_corp([3001]))

    def test_is_ally(self):
        self.assertFalse(self.killmail.is_alliance([2001]))
        self.assertFalse(self.killmail.is_alliance([2001]))
        self.assertTrue(self.killmail.is_alliance([3001]))

    def test_is_structure(self):
        self.killmail.victim_ship.eve_group.eve_category_id = 65
        self.assertTrue(self.killmail.is_structure())
        self.killmail.victim_ship.eve_group.eve_category_id = 1
        self.assertFalse(self.killmail.is_structure())

    def test_is_mobile(self):
        self.killmail.victim_ship.eve_group.eve_category_id = 22
        self.assertTrue(self.killmail.is_mobile())
        self.killmail.victim_ship.eve_group.eve_category_id = 1
        self.assertFalse(self.killmail.is_mobile())

    def test_is_capsule(self):
        self.killmail.victim_ship.id = 670
        self.assertTrue(self.killmail.is_capsule())
        self.killmail.victim_ship.id = 33328
        self.assertTrue(self.killmail.is_capsule())
        self.killmail.victim_ship.id = 1
        self.assertFalse(self.killmail.is_capsule())

    def test_get_month(self):
        self.killmail.killmail_date = datetime(2023, 10, 1)
        self.assertTrue(self.killmail.get_month(10))
        self.assertFalse(self.killmail.get_month(9))

    def test_attackers_distinct_alliance_ids(self):
        self.assertEqual(
            set(self.killmail.attackers_distinct_alliance_ids()), {3001, 3002}
        )

    def test_attackers_distinct_corporation_ids(self):
        self.assertEqual(
            set(self.killmail.attackers_distinct_corporation_ids()), {2001, 2002}
        )

    def test_attackers_distinct_character_ids(self):
        self.assertEqual(
            set(self.killmail.attackers_distinct_character_ids()), {1001, 1002}
        )

    def test_threshold(self):
        killmail = Killmail.objects.get(killmail_id=1)
        self.assertTrue(killmail.threshold(500_000))
        self.assertFalse(killmail.threshold(1_500_000))

    def test_attacker_main_empty(self):
        mains = {}
        killmail = Killmail.objects.get(killmail_id=1)
        main, eve_id, attacker = killmail.attacker_main(mains)
        self.assertEqual(main, None)
        self.assertEqual(eve_id, None)
        self.assertIsNone(attacker)

    def test_attacker_main(self):
        # Create mains data with an alt and associated main character
        main_character = EveCharacter.objects.get(character_id=1001)
        alt_character = EveCharacter.objects.get(character_id=1005)
        mains = {1001: {"main": main_character, "alts": [alt_character]}}
        killmail = Killmail.objects.get(killmail_id=1)
        main, eve_id, attacker = killmail.attacker_main(mains)
        self.assertEqual(main, EveCharacter.objects.get(character_id=1001))
        self.assertEqual(eve_id, 1001)
        self.assertIsNone(attacker)

    def test_attacker_main_alt(self):
        # Create mains data with an alt and associated main character
        main_character = EveCharacter.objects.get(character_id=1001)
        alt_character = EveCharacter.objects.get(character_id=1005)
        mains = {1001: {"main": main_character, "alts": [alt_character]}}
        killmail = Killmail.objects.get(killmail_id=2)
        main, eve_id, attacker = killmail.attacker_main(mains)
        self.assertEqual(main, EveCharacter.objects.get(character_id=1001))
        self.assertEqual(eve_id, 1005)
        self.assertEqual(attacker, EveCharacter.objects.get(character_id=1005))

    def test_str(self):
        self.assertEqual(str(self.killmail), "Killmail 1")

    def test_get_image_url(self):
        expected_url = eveimageserver._eve_entity_image_url(
            self.victim.category, self.victim.eve_id, 32
        )
        self.assertEqual(self.killmail.get_image_url(), expected_url)



from datetime import datetime

from django.test import TestCase

from killstats.templatetags.killstats import current_month


class CurrentMonthFilterTest(TestCase):
    def test_current_month_with_valid_month(self):
        """Testet, ob ein gültiger Monatswert korrekt gehandhabt wird."""
        month = current_month(3)
        self.assertEqual(month, "March")

    def test_current_month_with_invalid_month(self):
        """Testet, ob ein ungültiger Monatswert den aktuellen Monat zurückgibt."""
        month = current_month(14)
        self.assertEqual(month, datetime.now().strftime("%B"))



"""Generate AllianceAuth test objects from JSON data."""

import json
from datetime import datetime
from pathlib import Path

from eveuniverse.models import EveType

from allianceauth.eveonline.models import (
    EveAllianceInfo,
    EveCharacter,
    EveCorporationInfo,
)

from killstats.models import AlliancesAudit, CorporationsAudit, EveEntity, Killmail
from killstats.tests.testdata.load_eveuniverse import load_eveuniverse


def _load_eveentity_data():
    with open(Path(__file__).parent / "eveentity.json", encoding="utf-8") as fp:
        return json.load(fp)


_entities_data = _load_eveentity_data()


def _load_killmail_data():
    with open(Path(__file__).parent / "killmails.json", encoding="utf-8") as fp:
        return json.load(fp)


def _load_get_bulk_data():
    with open(Path(__file__).parent / "get_bulk.json", encoding="utf-8") as fp:
        return json.load(fp)


_killstats_data = _load_killmail_data()


def load_killstats_all():
    load_eveentity()
    load_eveuniverse()
    load_killstats()
    load_corporationaudit()
    load_allianceaudit()


def load_eveentity():
    EveEntity.objects.all().delete()
    for character_info in _entities_data.get("EveEntity"):
        EveEntity.objects.create(
            eve_id=character_info.get("eve_id"),
            name=character_info.get("name"),
            category=character_info.get("category"),
        )


def load_corporationaudit():
    CorporationsAudit.objects.all().delete()
    CorporationsAudit.objects.update_or_create(
        id=1,
        corporation=EveCorporationInfo.objects.get(corporation_id=2001),
        owner=EveCharacter.objects.get(character_id=1001),
    )
    CorporationsAudit.objects.update_or_create(
        id=2,
        corporation=EveCorporationInfo.objects.get(corporation_id=2002),
        owner=EveCharacter.objects.get(character_id=1002),
    )


def load_allianceaudit():
    AlliancesAudit.objects.all().delete()
    AlliancesAudit.objects.update_or_create(
        id=1,
        alliance=EveAllianceInfo.objects.get(alliance_id=3001),
        owner=EveCharacter.objects.get(character_id=1001),
    )
    AlliancesAudit.objects.update_or_create(
        id=2,
        alliance=EveAllianceInfo.objects.get(alliance_id=3002),
        owner=EveCharacter.objects.get(character_id=1002),
    )


def load_killstats():
    Killmail.objects.all().delete()
    for killmail in _killstats_data.get("Killmail"):
        Killmail.objects.update_or_create(
            killmail_id=killmail.get("killmail_id"),
            killmail_date=datetime.strptime(
                killmail.get("killmail_date"), "%Y-%m-%d %H:%M:%S"
            ),
            victim=EveEntity.objects.get(eve_id=killmail.get("victim")),
            victim_ship=EveType.objects.get(id=killmail.get("victim_ship")),
            victim_corporation_id=killmail.get("victim_corporation_id"),
            victim_alliance_id=killmail.get("victim_alliance_id", None),
            victim_total_value=killmail.get("victim_total_value"),
            victim_fitted_value=killmail.get("victim_fitted_value"),
            victim_destroyed_value=killmail.get("victim_destroyed_value"),
            victim_dropped_value=killmail.get("victim_dropped_value"),
            victim_region_id=killmail.get("victim_region_id"),
            victim_solar_system_id=killmail.get("victim_solar_system_id"),
            victim_position_x=killmail.get("victim_position_x"),
            victim_position_y=killmail.get("victim_position_y"),
            victim_position_z=killmail.get("victim_position_z"),
            hash=killmail.get("hash"),
            attackers=killmail.get("attackers"),
        )



# Generated by Django 4.2.11 on 2024-06-28 07:27

import django.core.validators
import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ("eveonline", "0017_alliance_and_corp_names_are_not_unique"),
        ("eveuniverse", "0010_alter_eveindustryactivityduration_eve_type_and_more"),
    ]

    operations = [
        migrations.CreateModel(
            name="General",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
            ],
            options={
                "permissions": (
                    ("basic_access", "Can access this app, Killstats."),
                    ("admin_access", "Has access to all killboards."),
                ),
                "managed": False,
                "default_permissions": (),
            },
        ),
        migrations.CreateModel(
            name="EveEntity",
            fields=[
                (
                    "eve_id",
                    models.IntegerField(
                        primary_key=True,
                        serialize=False,
                        validators=[django.core.validators.MinValueValidator(0)],
                        verbose_name="id",
                    ),
                ),
                (
                    "category",
                    models.CharField(
                        choices=[
                            ("alliance", "Alliance"),
                            ("corporation", "Corporation"),
                            ("character", "Character"),
                        ],
                        max_length=32,
                        verbose_name="category",
                    ),
                ),
                ("name", models.CharField(max_length=254, verbose_name="name")),
                ("last_update", models.DateTimeField(auto_now=True)),
                (
                    "alliance",
                    models.ForeignKey(
                        default=None,
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="alli",
                        to="killstats.eveentity",
                    ),
                ),
                (
                    "corporation",
                    models.ForeignKey(
                        default=None,
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="corp",
                        to="killstats.eveentity",
                    ),
                ),
            ],
            options={
                "default_permissions": (),
            },
        ),
        migrations.CreateModel(
            name="KillstatsAudit",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("last_update", models.DateTimeField(auto_now=True)),
                (
                    "corporation",
                    models.OneToOneField(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="eveonline.evecorporationinfo",
                    ),
                ),
                (
                    "owner",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="eveonline.evecharacter",
                    ),
                ),
            ],
            options={
                "default_permissions": (),
            },
        ),
        migrations.CreateModel(
            name="Killmail",
            fields=[
                (
                    "killmail_id",
                    models.PositiveBigIntegerField(primary_key=True, serialize=False),
                ),
                ("killmail_date", models.DateTimeField(blank=True, null=True)),
                ("victim_corporation_id", models.PositiveBigIntegerField()),
                (
                    "victim_alliance_id",
                    models.PositiveBigIntegerField(blank=True, null=True),
                ),
                ("hash", models.CharField(max_length=255, unique=True)),
                (
                    "victim_total_value",
                    models.PositiveBigIntegerField(blank=True, null=True),
                ),
                (
                    "victim_fitted_value",
                    models.PositiveBigIntegerField(blank=True, null=True),
                ),
                (
                    "victim_destroyed_value",
                    models.PositiveBigIntegerField(blank=True, null=True),
                ),
                (
                    "victim_dropped_value",
                    models.PositiveBigIntegerField(blank=True, null=True),
                ),
                (
                    "victim_region_id",
                    models.PositiveBigIntegerField(blank=True, null=True),
                ),
                (
                    "victim_solar_system_id",
                    models.PositiveBigIntegerField(blank=True, null=True),
                ),
                ("victim_position_x", models.FloatField(blank=True, null=True)),
                ("victim_position_y", models.FloatField(blank=True, null=True)),
                ("victim_position_z", models.FloatField(blank=True, null=True)),
                ("attackers", models.JSONField(blank=True, null=True)),
                (
                    "victim",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        to="killstats.eveentity",
                    ),
                ),
                (
                    "victim_ship",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        to="eveuniverse.evetype",
                    ),
                ),
            ],
            options={
                "default_permissions": (),
            },
        ),
    ]



from django.db.models import Q, Sum

from killstats.api.account_manager import AccountManager
from killstats.api.helpers import KillboardDate, get_alliances, get_corporations
from killstats.api.killboard_manager import killboard_dashboard, killboard_hall
from killstats.models.killboard import Killmail


# pylint: disable=too-many-locals
def get_killmails_data(
    request, month, year, entity_id: int, mode, page_size: int, entity_type: str
):
    if entity_id == 0:
        if entity_type == "alliance":
            entities = get_alliances(request)
        else:
            entities = get_corporations(request)
    else:
        entities = [entity_id]

    start = int(request.GET.get("start", 0))
    length = int(request.GET.get("length", page_size))
    search_value = request.GET.get("search[value]", "")
    order_column_index = int(request.GET.get("order[0][column]", 0))
    order_dir = request.GET.get("order[0][dir]", "desc")

    column_mapping = {
        0: "killmail_id",
        1: "victim_ship__name",
        2: "victim",
        3: "victim__name",
        4: "victim_total_value",
        5: "killmail_date",
    }

    order_column = column_mapping.get(order_column_index, "killmail_date")
    order_by = f"{'-' if order_dir == 'desc' else ''}{order_column}"

    offset = start
    limit = length

    killmails = (
        Killmail.objects.prefetch_related("victim", "victim_ship")
        .filter(killmail_date__year=year, killmail_date__month=month)
        .order_by(order_by)
    ).filter_entities(entities)

    if mode == "losses":
        killmails = killmails.filter_entities_losses(entities)
    else:
        killmails = killmails.filter_entities_kills(entities)

    totalvalue = (
        killmails.aggregate(totalvalue=Sum("victim_total_value"))["totalvalue"],
    )

    if search_value:
        killmails = killmails.filter(
            Q(victim__name__icontains=search_value)
            | Q(victim_ship__name__icontains=search_value)
        )

    total_count = killmails.count()
    record_count = killmails.count()
    killmails = killmails[offset : offset + limit]

    output = []
    for killmail in killmails:
        output.append(
            {
                "killmail_id": killmail.killmail_id,
                "killmail_date": killmail.killmail_date,
                "victim": {
                    "id": killmail.victim.eve_id,
                    "name": killmail.victim.name,
                },
                "victim_ship": {
                    "id": killmail.victim_ship.id,
                    "name": killmail.victim_ship.name,
                },
                "victim_corporation_id": killmail.victim_corporation_id,
                "victim_alliance_id": killmail.victim_alliance_id,
                "hash": killmail.hash,
                "victim_total_value": killmail.victim_total_value,
                "victim_fitted_value": killmail.victim_fitted_value,
                "victim_destroyed_value": killmail.victim_destroyed_value,
                "victim_dropped_value": killmail.victim_dropped_value,
                "victim_region_id": killmail.victim_region_id,
                "victim_solar_system_id": killmail.victim_solar_system_id,
                "victim_position_x": killmail.victim_position_x,
                "victim_position_y": killmail.victim_position_y,
                "victim_position_z": killmail.victim_position_z,
                "attackers": killmail.attackers,
            }
        )

    return {
        "draw": int(request.GET.get("draw", 1)),
        "recordsTotal": total_count,
        "recordsFiltered": total_count if not search_value else record_count,
        "data": output,
        "totalvalue": totalvalue,
    }


def get_killstats_halls(request, month, year, entity_id: int, entity_type: str):
    if entity_id == 0:
        if entity_type == "alliance":
            entities = get_alliances(request)
        else:
            entities = get_corporations(request)
    else:
        entities = [entity_id]

    if entity_type == "alliance":
        account = AccountManager(alliances=entities)
    else:
        account = AccountManager(corporations=entities)

    killmails = (
        Killmail.objects.prefetch_related("victim", "victim_ship").filter(
            killmail_date__year=year,
            killmail_date__month=month,
        )
    ).filter_entities(entities)

    mains, _ = account.get_mains_alts()

    shame, fame = killboard_hall(killmails, entities, mains)

    halls = []
    halls.append(
        {
            "shame": shame,
            "fame": fame,
        }
    )

    return halls


def get_killstats_stats(request, month, year, entity_id: int, entity_type: str):
    if entity_id == 0:
        if entity_type == "alliance":
            entities = get_alliances(request)
        else:
            entities = get_corporations(request)
    else:
        entities = [entity_id]

    killmail_year = (
        Killmail.objects.prefetch_related("victim", "victim_ship")
        .filter(killmail_date__year=year)
        .order_by("-killmail_date")
    ).filter_entities(entities)

    date = KillboardDate(month, year)

    stats = killboard_dashboard(killmail_year, date, entities)

    output = []
    output.append(
        {
            "stats": stats,
        }
    )

    return output
